// --- [ DCF_77_mini.c ] ------------------------------------------------------
//       29.09.2011
//       tab = 3
//			Auf die Uhrfunktion reduziert Version ohne Ausgabe
//       Hat die Aufgabe, die interne Uhr eines Datenloggers zu steuern.
// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "dcf_77_mini.h"

// --- Konstanten -------------------------------------------------------------

#define TRUE			1
#define FALSE			0

#define MS_100			9						// Counts fr 0.1 Sek Puls [Einheit 0.01sec]
#define MS_200			19						// Counts fr 0.2 Sek Puls
#define MS_1000		99						// Counts fr 1.0 Sek Puls
#define MS_2000		199					// Counts fr 2.0 Sek Puls
#define TOL1			3						// zulssige Zeitabweichung +- des DCF-Signals (bei 0.1/0.2 Sekunden)
#define TOL2			7						// zulssige Zeitabweichung +- des DCF-Signals (bei 1/2 Sekunden)
													// jeweils das 6. Element in den TimeArrays:
													// time[CLK][x] - die aktuelle Sekunde
#define ZBITS			6						// time[DCF][x] - die Zusatzbits des DCF-Signal
#define TIMEBYTES		7						// Anzahl der Bytes in jedem Timearray

													// Bitpositionen fr 'flag'
#define SYNCH_ENABLE 0						// zeigt dan, dass eine Synchronisation mit der DCF-Zeit gefordert ist
#define XTL_SEC		1						// neue Sekunde ist angebrochen
#define DCF_RX_58		2						// 58 DCF-Bits empfangen
#define XTL_MINUTE	3						// neue Minute der internen xtal-Uhr erreicht
#define DCF_MINUTE	4						// neue Minute aus DCF-Signal empfangen
#define DISPLAY		5						// die Anzeige muss aktualisiert werden
#define SYNCH			6						// Die Zeit wurde neu auf die DCF-Zeit synchronisiert
#define START_FLAG	7						// Signalisiert den Programmstart, bentigt zur bernahme der 1. Minute

#define M168									// Register fr den ATMEGA168/88/48
//#define M8

													// Bytepositionen im TimeArray
#define CLK				0						// die fr die Anzeige massgebliche Zeit der Uhr (+ 1Byte Sekunde)
#define DCF				1						// die Zeit des DCF-Signals
#define BK2				2						// eine Kopie der vorletzten Minute
#define CMP				3						// die um 1 Minute incremenierte Zeit fr Plausibiltsprfung
#define BAK				4						// eine Kopie der letzten Minute fr Plausibilittsprfung
#define BK1				4						// ein anderer Name fr BAK -> BK1 und BK2
#define CLK_ARRAYS	5						// die Anzahl der Arrays von CLK bis BAK

#define ERR_PARITY	0						// Parittsfehler
#define ERR_BIT_CNT	1						// ungleich 58 Bit empfangen
#define ERR_SIGNAL	2						// DCF_Signal-Fehler (Counter berlauf)
#define ERR_NO_DCF	3						// Signalabstand grer als 2 Sekunden
#define ERR_BIT_CNT2 4						// mehr als 58 Bit empfangen
#define ERR_TWI   	7						// Die Uhrzeit wurde zum Debuggen via TWI gendert

#define SYNCH_TIME	14						// Die Anzahl der Minuten, die die Uhr ohne Synchronisation mit dem DCF-Signal
													// nur nach dem XTL-Takt luft
													
#define TIMER1			1						// Timer0 ist fest vorgegeben, zwischen Timer1 und Timer2 kann gewhlt werden
#define TIMER2			2
#define TIMER 			TIMER1				// legt fest, ob der Timer1 oder der Timer2 als 2. Timer benutzt wird

#define LED_PORT		PORTB
#define LED_PIN		PINB
#define LED_DDR		DDRB
#define LED_1_PIN		PB0

#define LED_CONFIG	(LED_DDR |= (1<<LED_1_PIN))

#define LED_1_OFF		(PORTB |=  (1<<LED_1_PIN))
#define LED_1_ON		(PORTB &= ~(1<<LED_1_PIN))
#define LED_1_TOG		(LED_PIN = (1<<LED_1_PIN))

// --- Variablen --------------------------------------------------------------

uint8_t summand[] 		PROGMEM = { 1, 2, 4, 8, 10, 20, 40, 80 }; // fr die Decodierung des DCF77-Signals
uint8_t days_of_month[] PROGMEM = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
//char day_names[] 			PROGMEM = {'M','o','D','i','M','i','D','o','F','r','S','a','S','o' };

volatile uint8_t flag;						// sammelt Status-Flags
//volatile uint8_t flag2;						// ein weiteres Flagbyte fr den Beginn der 59. Sekunde
volatile uint8_t error;						// sammelt error-Flags
volatile uint8_t T0_ticks;					// Counter fr die interne Uhr / Sekundensignal
volatile uint8_t T1_ticks;					// Counter fr die Lnge des DCF-Signals (in 0.01sek)
uint8_t error_abs;							// summiert die Anzahl der DCF-Fehler
													// wird um 0.00h zurckgesetzt
uint8_t DCF_valid = TRUE;
uint8_t synch_time = 0;						// damit nach dem Programmstart eine Synchronisation erzwungen wird

uint8_t time[CLK_ARRAYS][TIMEBYTES];   // Arrays fr die aktuelle Zeit sowie Kopien davon (7 Arrays)

uint8_t dcf_bits[8];							// lokale Variable fr 64 DCF-Bits
uint8_t bit_idx = 0;							// lokaler Zhler fr DCF-Bit-Nummer

//char cv[4];										// fr utoa() falls serielle Ausgabe erforderlich
uint8_t clock_start = TRUE;				// Signalisiert den Start der Uhr ohne erfolgte Synchronisation

// --- Prototypen -------------------------------------------------------------

void 		DCF_decode(void);
void 		DCF_decr_time(uint8_t *time, uint8_t minute);
void 		DCF_incr_time(uint8_t *time, uint8_t minute);
void 		DCF_copy_time(uint8_t *target, uint8_t *source);
uint8_t 	DCF_compare_time(uint8_t *time1, uint8_t *time2);
void 		DCF_backup(void);
uint8_t 	DCF_time_valid(void);
void    	DCF_save_bit(uint8_t v);
void    	DCF_reset_bits(void);
uint8_t 	DCF_get_bit(uint8_t bit);
uint8_t 	DCF_bits_received(void);

uint8_t 	is_leap(uint8_t j);
void 		switch_summertime(uint8_t *time);
uint8_t 	is_summertime(uint8_t *time);
void 		set_summertime(uint8_t *time, uint8_t i);
uint8_t 	calc_time(uint8_t i, uint8_t cnt);
uint8_t 	calc_parity(uint8_t i, uint8_t cnt);


// --- PORTs ------------------------------------------------------------------
// PD.0/1	RxD, TxD 						// fr serielle Schnittstelle beim Debuggen
// PD.2/3	INT0, INT1						// hier wird die Uhr angeschlossen
// PB.6/7	XTAL								// der Quarz bentigt zwei Pin
// PB.0     fr LED

// --- DCF-Empfnger ----------------------------------------------------------
// erwartet wird vom DCF-Empfnger:
// DCF-Ruhe-Level: HIGH, bei Signal: LOW !!!!!!!!!!!!!!!!!!!!!!!!!

// --- Timer ------------------------------------------------------------------
// es werden 2 Timer (8 oder 16 Bit) verwendet.

// ----------------------------------------------------------------------------
void DCF_Init(void)
{
// --- EXT_INT ----------------------------------------------------------------
#ifdef M168
EICRA = (1<<ISC11 | 1<< ISC10 | 1<< ISC01);// Rising INT1, Falling INT0
EIMSK  = (1<<INT1 | 1<< INT0);			// INT0 / INT1_isr enablen
#endif
#ifdef M8
MCUCR = (1<<ISC11 | 1<< ISC10 | 1<< ISC01);// Rising INT1, Falling INT0
GICR  = (1<<INT1 | 1<< INT0);			// INT0 / INT1_isr enablen
#endif
// --- CLock-Prescaler --------------------------------------------------------
#ifdef M168
CLKPR = (1<<CLKPCE);
CLKPR &= 0xF0;									// lower nibble lschen -> Faktor 1
#endif

// --- TIMER0 -----------------------------------------------------------------
// generiert den internen 1-Sekunden-Takt
// 3.686.400 / ((64 * 256) * 225) = 1 Sekunde
TCCR0B = (1<<CS01 | 1<< CS00);			// Prescale 64, den Rest macht TIMER0_ISR
#define T0_PRELOAD	225

#if TIMER == TIMER1
// --- TIMER1 -----------------------------------------------------------------
// luft als Zhler fr Pulsdauer und Pulsabstand
// 3.686.400 / (256 * 144)`= 0.01 Sekunde
TCCR1A = 0;
TCCR1B = (1<<CS12);							// Prescale 256
OCR1AH = 0;
OCR1AL = 144;

#elif TIMER == TIMER2
// --- Timer2 -----------------------------------------------------------------
// luft als Zhler fr Pulsdauer und Pulsabstand
// 3.686.400 / (256 * 144)`= 0.01 Sekunde
TCCR2A = 0;
TCCR2B = ((1<<CS22) | (1<<CS21) | (0<<CS20));	// Prescale 256
OCR2A  = 144;
#else
	#error 2. Timer fehlt
#endif

// --- beide Timer ------------------------------------------------------------
#ifdef M8
//TIMSK = (1<<TOIE0 | 1<<OCIE1A);		// ISR fr Timer1 / 2 enablen
#endif
#ifdef M168
TIMSK0 = (1<<TOIE0 );						// ISR fr Timer0 enablen
#if TIMER == TIMER1
TIMSK1 = (1<<OCIE1A);						// ISR fr Timer1 enablen
#elif TIMER == TIMER2
TIMSK2 = (1<<OCIE2A);
#endif
#endif

// --- PORTS ------------------------------------------------------------------
LED_CONFIG;
LED_1_ON;

flag = 0;
}

// ---------------------------------------------------------------------------
// liefert TRUE zurck, wenn der Beginn einer neuen Sekunde oder einer
// neuen Minute erkannt wurde.
// Danach muss die Funktion DCF77_Check_Time() aufgerufen werden,
// damit die Zeit fortgeschrieben wird 
uint8_t DCF_New_Second(void)
{
return (flag & ( (1<<XTL_SEC) | (1<<DCF_MINUTE) | (1<< XTL_MINUTE)) );
}

// ----------------------------------------------------------------------------
// Liefert die angefragte Zeitkomponente als Zahl zurck
uint8_t DCF_Get_Time(uint8_t idx)
{
return(time[CLK][idx]);
}

// ----------------------------------------------------------------------------
// Liefert die angefragte Zeitkomponente als String 2-stellig zurck
void DCF_Get_TimeStr(uint8_t idx, char * s)
{
if ((time[CLK][idx]) < 10) 
	{
	s[0] = '0';
	s[1] = time[CLK][idx] + '0';
	s[2] = 0;
	}
else utoa(time[CLK][idx], s, 10);
}

// ----------------------------------------------------------------------------
// Liefert die Anzahl der DCF-Fehler zurck, die seit dem Programmstart
// bzw. dem letzten Aufruf der Funktion registriert wurden.
// Im normalen Betrieb sind es nicht mehr als ca. 3 Fehler am Tag.
// Lediglich bei einem Gewitter kann die Anzahl deutlich hher werden.

uint8_t DCF_Get_Errors(void)
{
	uint8_t e;
	e = error_abs;
	error_abs = 0;
	return(e);
}
// ----------------------------------------------------------------------------
// Eine neue Sekunde oder gar eine neue Minute hat begonnen, 
// hier nachsehen, was zu tun ist ....
void DCF_Check_Time()
{
	if (clock_start)								// Es ist noch keine Synchronisation mit dem Minutenanfang erfolgt !
	{
		if (flag & (1<<DCF_MINUTE))
		{
														// Flag fr Neustart und Neusychronisation setzen
			flag 		= ( 1<<START_FLAG | 1<<SYNCH_ENABLE );
			clock_start = FALSE;					// Der Minutenbeginn ist gefunden.
			DCF_reset_bits();						// Den Index von DCF_bits auf 0 setzen
			time[CLK][SEKUNDE] = 0;				// Bei Sekunde 0 starten.
			error 	= 0;							// Alle Fehler vergessen.
			T1_ticks = 0;							// Alle Counter zurcksetzen

#if TIMER == TIMER1
			TCNT1H   = 0;
			TCNT1L   = 0;
#elif TIMER == TIMER2
			TCNT2    = 0;
#else
	#error 2. Timer fehlt !
#endif
		}
		
		if (flag & (1<<XTL_SEC))				// wenn das Sekundenflag gesetzt ist
		{
			flag &= ~(1<<XTL_SEC);				// damit nicht 60 Sekunden erreicht werden
			if (time[CLK][SEKUNDE] == 59) time[CLK][SEKUNDE]--;
		}
	}
	
	else												// der Minutenanfang ist bereits erkannt
		{
		if (error) DCF_valid = FALSE;			// sobald ein Fehler aufgetreten ist,
														// wird auf die xtal-Uhr umgeschaltet.

// --- neue Sekunde der xtal-Uhr ----------------------------------------------

		if (flag & (1<<XTL_SEC))				// das Flag fr eine neue Sekunde ist gesetzt
		{
			flag &= (~(1<<XTL_SEC));
	
			if (!(time[CLK][TAG])) LED_1_TOG;// Solange noch kein Tag definiert ist, LED togglen
			else if (error) LED_1_ON;			// danach nur einschalten, wenn ein Fehler aufgetreten ist

														// Bei einem Minutenwechsel werden zwei Flags gesetzt,
														// XTL_MINUTE und DCF_MINUTE, die Reihenfolge ist zufllig.
														// Um zu verhindern, dass nach dem ndern von SYNCH_ENABLE
														// beide zuschlagen und die Zeit 2x incrementiert wird, wird
			if (time[CLK][SEKUNDE] == 55)		// das Flag erst so spt wie mglich gendert.
			{								
				if (synch_time)
				{										// Das Zeitintervall bis zum nchsten Synchronisationsversuch
					synch_time--;					// decrementieren - bei 0 wird synchronisiert
					flag &= (~(1<<SYNCH_ENABLE));	// das Flag zur Synchronisation lschen
				}
				else
				{
					flag |= (1<<SYNCH_ENABLE);	// bei 0 das Flag setzen
														// wenn das Flag DCF_MINUTE nach dem Flag XTL_MINUTE gesetzt wird,
														// kann es einen zustzlichen Minutenanfang auslsen, da jetzt SYNCH_ENABLE
														// gesetzt ist.
														// Um das zu verhindern, wird das Flag prophylaktisch gelscht und 
														// DCF_reset_bits() aufgerufen (was normalerweise durch DCF_MINUTE ausgefhrt wird)
				}
			}
		}

// --- In der XTAL-Zeit ist eine neue Minute angebrochen
// --- wenn DCF_valid == FALSE, dann ist die XTL-Zeit fr die Anzeige mageblich
// --- ausserdem wird diese Zeit verwendet, wenn keine Synchronisation mit der DCF-Zeit
// --- gefordert ist.

		if (flag & (1<<XTL_MINUTE))
		{
			flag &= (~(1<<XTL_MINUTE));		// das Flag lschen

			if (!(DCF_valid) || !(flag & (1<<SYNCH_ENABLE)))
			{
				DCF_incr_time(&time[CLK][0], 1);
				time[CLK][SEKUNDE] = 0;
				flag |= (1<<DISPLAY);			// Flag veranlasst die Anzeige der Zeit
			}
		}

// --- In der DCF-Zeit wurde der Minutenbeginn erkannt
// --- wenn DCF_valid  == TRUE vorliegen, dann ist diese Zeit mageblich
// --- es sei denn, dass das Flag SYNCH_ENABLE nicht gesetzt ist

		if (flag & (1<<DCF_MINUTE))
		{
			flag &= (~(1<<DCF_MINUTE));	// das Flag lschen
													// wenn kein grundstzlicher Fehler vorliegt
			if ( (DCF_valid) && (flag & (1<<SYNCH_ENABLE)) )	// und das Flag SYNCH_ENABLE gesetzt ist				
			{
													// wenn aber nur ein unvollstndiger Bit-Satz vorliegt, kann kein Fehler
				if(!(DCF_bits_received() == 58))// erkannt worden sein, daher hier noch eine Prfung der DCF-Bits.
				{									// DCF_valid bergibt im Fehlerfall die Kontrolle an die xtal-Uhr.
					DCF_valid = FALSE;			// Wenn die Sekunden > 59, dann wurde das Flag XTL_MINUTE bereits gesetzt,
													// wegen DCF_valid hat es aber keine Aktion ausgelst.
													// daher Sekunden prfen und ggf. das Flag noch einmal setzen.
													// zhlen die Sekunden weniger als 60, dann schlgt das Flag via Xtal-Zeit
													// von sich aus zu.
					if (time[CLK][SEKUNDE] > 59) flag |= (1<<XTL_MINUTE);
				}

				else 									// wenn alle DCF-Bits empfangen wurden
				{
					synch_time = SYNCH_TIME;	// wieder auf den Vorgabewert einstellen
													// synch_time muss > 0 sein, wenn das Flag SYNCH_ENABLE gelscht werden soll
													// Das Flag wird erst bei der ANzeige der Minute gelscht, damit nicht 
													// eine eventuell folgendes Flag XTL_MINUTE noch einmal die Zeit incrementiert..

													// Wenn kein Fehler vorlag, xtal-Uhr auf Minutenbeginn
					TCNT0 = 0;						// synchronisieren
					T0_ticks = T0_PRELOAD;		// Timer0 decrementiert T0_ticks auf 0 fr 1 Sekunde
					time[CLK][SEKUNDE] = 0;
					DCF_incr_time(&time[CLK][0], 1);

													// wenn die XTL- und DCF-Zeiten nicht bereinstimmen,
													// hier die DCF-Zeit bernehmen
					if (!(DCF_compare_time(&time[CLK][0], &time[DCF][0])))
					{
						DCF_copy_time(&time[CLK][0], &time[DCF][0]);
						flag |= (1<<SYNCH);	// Nur fr das Debuggen Hinweis auf Neusychronisation setzen
					}								// nicht in error ablegen, sonst wird DCF_valid auf FALSE gesetzt
													// und es wird ein weiterer Minutenbeginn durch die xtal-Zeit mglich
					flag |= (1<<DISPLAY);	// Flag veranlasst die Anzeige der Zeit
				}
			}
		DCF_reset_bits();						// Die DCF-bits zurcksetzen/auf den Minutenbeginn synchronisieren
													// Darf erst nach der letzten Auswertung von DCF_bit_received() erfolgen !
													// Darf ausschlielich hier nach einem Flag DCF_Minute ausgefhrt werden !
		}

// --- Nach 58 eingegangenen DCF-Bits wird das Signal decodiert und auf Fehler geprft
// --- nur wenn der Vergleich mit den vorausgegangenen 2 Minuten plausibel ist, wird DCF_valid auf TRUE gesetzt
// --- dumm ist es nur, wenn DCF-Bits 'verschluckt' wurden, dann ist das Bit DCF_RX_58 zum Ende der Minute nicht gesetzt

		if (flag & (1<<DCF_RX_58))			// das Flag wird in DCF_save_bit() gesetzt, wenn 58 Bit eingegangen sind
		{
			flag &= (~(1<<DCF_RX_58));		// das Flag lschen
			DCF_decode();						// nun steht die decodierte Zeit der neuen Minute in time[DCF]
													// in time[CLK] steht noch die aktuelle Minute
													// error kann noch durch Parittsfehler von DCF_decode() gesetzt werden
													// oder durch fehlerhafte Impulslngen whrend der letzten Minute

													// die aktuelle Zeit mit den letzten beiden DCF-Minuten vergleichen
			if ( (DCF_time_valid()) &&  (!(error)) ) DCF_valid = TRUE; else DCF_valid = FALSE;
			if (!(error)) DCF_backup();	// legt ein Backup der letzten beiden Minuten an

													// eine Ausnahme, damit der Neustart schneller vonstatten geht
													// das Startflag ist nach einem Neustart gesetzt - und nur dann !!
			if ((flag & (1<<START_FLAG)) && (!(error)))
			{
				flag &= ~(1<<START_FLAG);
				DCF_valid = TRUE;				// das DCF-Signal setzt den Minutenanfang
													// und kopiert die DCF-Zeit in die Uhr
			}
		}

		
// --- Minuten-Anzeige aktualisieren ------------------------------------------
// --- allerdings nur dann, wenn nicht noch das Flag DCF_Minute ansteht
// --- das Flag muss vorher abgearbeitet werden
//
// --- Eine Anzeige gibt es in diesem Modul nicht, dennoch mssen einige
//     Einstellungen am Minutenbeginn vorgenommen werden
		if ((flag & (1<<DISPLAY))  && (!(flag & (1<<DCF_MINUTE))))
		{
/*		
			put_s(utoa(time[CLK][STUNDE], cv, 10));
			put_c(':');
			put_s(utoa(time[CLK][MINUTE], cv, 10));
			put_c(':');
			put_s(utoa(time[CLK][SEKUNDE], cv, 10));
			put_c(' ');
			put_s(utoa(time[CLK][TAG], cv, 10));
			put_c('.');
			put_s(utoa(time[CLK][MONAT], cv, 10));
			put_c('.');
			put_s(utoa(time[CLK][JAHR], cv, 10));
			
			put_s(" err: 0x");
			put_s(utoa(error,cv,16));
						put_s( " DCF_valid: 0x");
			put_s(utoa(DCF_valid,cv,16));
									put_s( " Flag: 0x");
			put_s(utoa(flag,cv,16));
			put_c(13);
*/			
														// wegen der Timing-Problem erhlt das Flag DCF_MINUTE eine chance:
		if (flag & (1<<DCF_MINUTE))			// Falls dieses Flag vom DCF-Signal zwischenzeitlich 
		{												// noch gesetzt wurde, dann 
			flag &= (~(1<<DCF_MINUTE));		// das Flag lschen
			DCF_reset_bits();						// die DCF_bits reseten
		}												// verhindert, dass in der Folgeminute zu wenig DCF-Bits empfangen werden.
		
		error_abs += error;						// Anzahl der Fehler in error_abs aufsummieren.
		error = 0;									// Flags und Fehlerflags der letzten Minute zurcksetzen
		LED_1_OFF;									// Die LED zum Minutenbeginn immer abschalten

		flag &= (~(1<<DISPLAY | 1<<SYNCH));	// NUR die Flags Display und SYNCH lschen, aber erst nach der Aktualisierung des Displays / Log-Ausgabe
		}												// falls das Flag DCF_MINUTE zwischenzeitlich noch gesetzt wurde, dann wird es noch abgearbeitet
	}
}


// -----------------------------------------------------------------------------
// INT1 - rising edge = Ende eines Impulses
// Es ist die Dauer des Impulses zu bestimmen (0.1 oder 0.2 Sekunden)
// schreibt das empfangene Bit direkt in das Bitarray.

ISR(INT1_vect)
{
register uint8_t x;
x = T1_ticks;

if      (x <  (MS_100 - TOL1)) return;// krzer als 0.1 Sekunden
													// 0.1 Sekunden
else if (x <  (MS_100 + TOL1)) DCF_save_bit(0); // eine 0 speichern

else if (x <  (MS_200 - TOL1)) return;	// krzer als 0.2 Sekunden
													// 0.2 Sekunden
else if (x <  (MS_200 + TOL1)) DCF_save_bit(1); // eine 1 speichern

}

// -----------------------------------------------------------------------------
// INT0 - falling edge = Start des Impulses
// Es wird geprft, wieviel Zeit seit dem letzten Impulsbeginn vergangen ist,
// es knnen 1 Sekunde (normales DCF_Bit) oder 2 Sekunden (Minutenbeginn) sein.
// Wesentliche Aufgabe des Interrupts ist, den 2-Sekunden-Signalabstand zu Beginn
// einer neuen Minute zu erkennen und den Beginn einer neuen Minute an main() zu signalisieren.
// Sind Empfangsfehler im DCF-Signal aufgetreten, dann bernimmt die XTL-Uhr die Aufgabe,
// den Beginn einer neuen Minute zu definieren.

// Hat auch die Aufgabe, beim Programmstart den Beginn der Zhlung zu starten.
ISR(INT0_vect)
{
register uint8_t x;
x = T1_ticks;									// volatile in Register laden

if      (x <  (MS_1000 - TOL2)) return;// krzer als 1 Sekunde ignorieren

else if (x <  (MS_1000 + TOL2));			// Normalfall 1 Sekunde - nur Counter aktualisieren

else if (x <  (MS_2000 - TOL2)) return;// krzer als 2 Sekunden ignorieren

else if (x <  (MS_2000 + TOL2))			// 2 Sekunden
	{
	flag  |= (1<<DCF_MINUTE );				// Flag fr neue DCF-Minute
	}

else error |= (1<< ERR_NO_DCF);			// lnger als 2 Sekunden, damit ungltig

T1_ticks = 0; 			// Counter fr Timer1 zurcksetzen
#if TIMER == TIMER1
	TCNT1H   = 0;
	TCNT1L   = 0;
#elif TIMER == TIMER2
	TCNT2    = 0;
#endif
}

// -----------------------------------------------------------------------------
// wird als interne Zeitreferenz fr den Sekundentakt genutzt
// Prescaler = 64, Preload = 256, Counter 225, 3686800 / (64 * 256 * 225) = 1 Sekunde
// Das Flag 1<<XTL_SEC  wird nach genau 1 Sekunde gesetzt und wird in main() ausgewertet.
// Bei fehlerfreiem DCF-Signal signalisiert das 2-Sek-DCF-Signal den Minutenstart.
// Bei fehlerhaftem DCF-Signal signalisiert die 60. Sekunde des Timer0 den Minutenstart.

ISR(TIMER0_OVF_vect)
{

if (--T0_ticks) return;						// abbrechen solange T0_ticks nicht zu 0 decrementiert wurde
T0_ticks = T0_PRELOAD;						// neuen Startwert setzen

time[CLK][SEKUNDE]++;						// Sekunde incrementieren

if (time[CLK][SEKUNDE] < 60)				// solange < 59 wird nur die Sekunde angezeigt
	{												// hier gibt der Timer den Takt an
	flag |= (1<<XTL_SEC);					// Flag fr neue Sekunde setzen
	}

else 												// (time[CLK][SEKUNDE] == 60)
	{
	flag |= (1<<XTL_MINUTE);				// Flag fr neue Minute durch CLK setzen
	}
}

#if TIMER == TIMER1
// -----------------------------------------------------------------------------
// fr die Bestimmung der Impulslngen, wird alle 0,01 Sekunden ausgelst
// bei normaler Funktion werden 100/200 ticks ausgelst
// kommt kein Signal, dann luft ticks ber, in diesem Fall wird der Wert
// konstant auf 255 gehalten.
// An diesem Wert kann erkannt werden, dass ein DCF-Signal berfllig ist.
ISR(TIMER1_COMPA_vect)
{
TCNT1H = 0;
TCNT1L = 0;
T1_ticks++;										// incrementieren bis zum berlauf auf 0
if (!(T1_ticks))
	{
	T1_ticks--;									// wenn T1_ticks berluft, dann zurcksetzen auf 255
	error |= (1<< ERR_SIGNAL);				// fehlendes DCF-Signal
	}
}

#elif TIMER == TIMER2
// -----------------------------------------------------------------------------
// fr die Bestimmung der Impulslngen, wird alle 0,01 Sekunden ausgelst
// bei normaler Funktion werden 100/200 ticks ausgelst
// kommt kein Signal, dann luft ticks ber, in diesem Fall wird der Wert
// konstant auf 255 gehalten.
// An diesem Wert kann erkannt werden, dass ein DCF-Signal berfllig ist.
ISR(TIMER2_COMPA_vect)
{
TCNT2 = 0;
T1_ticks++;										// incrementieren bis zum berlauf auf 0
if (!(T1_ticks))
	{
	T1_ticks--;									// wenn T1_ticks berluft, dann zurcksetzen auf 255
	error |= (1<< ERR_SIGNAL);				// fehlendes DCF-Signal
	}
}
#else
	#error 2. Timer fehlt
#endif

// ----------------------------------------------------------------------------
// Dekodiert die gelesenen 58 Bits und schreibt die Daten der nchsten Minute
// ins SRAM in das Array 'DCF_time[]'.
// Die zugehrigen Parittsbits werden berprft, Fehler werden in error markiert.
// Die Ankndigungs- und Zusatzbits werden in DCF_time[ZBITS] abgelegt.
// Die Ankndigungsbit stehen im Signal ab Pos. 15 in der Reihenfolge:
// R, A1, Z1, Z2, A2, S  (wobei S immer = 1).
// Sie werden umsortiert (warum eigentlich ?):
// DCF_zbits:(..|..|..|A1|A2|R |Z1|Z2)
// Sommerzeit(..|..|..|A1|A2|R | 1| 0)
// Winterzeit(..|..|..|A1|A2|R | 0| 1)
// A1 Ankndigungsbit fr Sommerzeit
// A2 Ankndigungsbit fr Schaltsekunde
// R  Reserveantenne aktiv
// Die Sommerzeit wird direkt als Bit.7 in time[CLK][DOW] abgelegt.

void DCF_decode(void)
{
													// zuerst die Zusatzbits nach DCF_time[ZBITS] schreiben
time[DCF][ZBITS]  = DCF_get_bit(16);			// A1
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(19);			// A2
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(15);			// R
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(17);			// Z1
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(18);			// Z2

time[DCF][MINUTE] = calc_time(21, 21 + 7);// Minute auslesen, Paritt prfen
if (calc_parity(21, 21 + 8)) error |= (1<<ERR_PARITY);

time[DCF][STUNDE] = calc_time(29, 29 + 6);// Stunde auslesen, Paritt prfen
if (calc_parity(29, 29 + 7)) error |= (1<<ERR_PARITY);

time[DCF][TAG] = calc_time(36, 36 + 6);// Tag auslesen

time[DCF][DOW] = calc_time(42, 42 + 3);// Day of week auslesen
													// die Nr. des Tag steht nun als Ziffer in DOW
													// z.B. Mittwoch = 3
													// umwandeln in Bitposition 0 ... 6
													// dazu die 1 um (3-1) mal nach links shiften
time[DCF][DOW] = 1 << (time[DCF][DOW] - 1);

													// die Sommerzeit als Bit.7 in den DOW schmuggeln
if ((time[DCF][ZBITS] & 0x03) == 2) set_summertime(&time[DCF][0], TRUE);

time[DCF][MONAT] = calc_time(45, 45 + 5);// Monat auslesen

time[DCF][JAHR] = calc_time(50, 50 + 8);// Jahr auslesen, Paritt prfen
if (calc_parity(36, 36 + 23)) error |= (1<<ERR_PARITY);
}

// -----------------------------------------------------------------------------
// Decodiert eine Zahl aus einem Bit-Bereich des DCF_Bit_Arrays,
// das ergibt dann den Wert fr entweder: Minute, Stunde, Tag, Monat, Jahr
// Parameter: Startposition des ersten Bits
//		        Endposition   des letzten Bits
// Rckgabe:  der ermittelte Wert
uint8_t calc_time(uint8_t start, uint8_t end)
{
// uint8_t summand[] PROGMEM = { 1, 2, 4, 8, 10, 20, 40, 80 };
// um PROGMEM zu nutzen, darf das Array nicht lokal sein

uint8_t r = 0;
uint8_t i = 0;

while (start < end)
	{
	if (DCF_get_bit(start++))
		{
		r += pgm_read_byte(&summand[i]);
		}
	i++;
	}

return(r);
}

// -----------------------------------------------------------------------------
// Berechnet die Paritt, beginnend an Position "start" ber "cnt" Bits
// (Achtung:  das letzte Bit im Bereich ist das Parittsbit !)
// Parameter: Startpunkt ab Arraybeginn
//		        Letzes Bit = Parittsbit
// Rckgabe:  (0 = kein Parittsfehler, 1 = Parittsfehler)
// Das Parittsbit ergnzt die Summe der einzelnen Bits auf eine gerade Zahl,
// d.h. Bit.0 ist nicht gesetzt - sonst liegt ein Fehler vor.
uint8_t calc_parity(uint8_t start, uint8_t end)
{
register uint8_t p = 0;
while (start < end)
	{
	if (DCF_get_bit(start++)) p++;
	}

return(p & 0x01);
}

// -----------------------------------------------------------------------------
// Ein Zeitarray wird um x Minuten incrementiert.
// ES wird die Zeitzhlung aktualisiert und incl. aller mglichen berlufe.
// Weiterhin erfolgt Prfung auf Wechsel zwischen Sommer-/Winterzeit.
void DCF_incr_time(uint8_t *zeit, uint8_t x)
{
register uint8_t t;

do	{
	zeit[MINUTE]++;
	if (zeit[MINUTE] < 60) continue;		// solange keine volle Stunde angebrochen ist,
													// kann hier beendet werden
	zeit[MINUTE] = 0;
	zeit[STUNDE]++;
													// zu jeder vollen Stunde testen, ob eine Wechsel
	switch_summertime(zeit);				// zwischen Sommer- und Winterzeit ansteht.
													// im jeweiligen Zeitarray wird dann das Flag fr
													// die Sommerzeit (DOW.7) gegebenfalls gendert
													// wenn Stunde == 24, dann kann kein Fehler entstehen,
													// da nur um 2 bzw. 3 Uhr umgeschaltet wird
	if (zeit[STUNDE] < 24) continue;
	zeit[STUNDE] = 0;
	zeit[TAG]++;
													// den Wochentag incrementieren
	t = zeit[DOW] & 0x80;					// Bit.7 im Wochentag markiert die Sommerzeit, zwischenspeichern !
	zeit[DOW] &= 0x7F;						// Bit.7 entfernen
	zeit[DOW] *=2;								// Wochentag incrementieren (nach links schieben)
													// berlauf von Sonntag auf Montag
	if (zeit[DOW] == 0x80) zeit[DOW] = 0x01;
	zeit[DOW] += t;							// Sommerzeit als Bit.7 wieder addieren

													// nun kommen die Tage des Monats in Spiel ...
	if (zeit[MONAT] == 2)					// zuerst Sonderfall Februar und Schaltjahr
		{
		if (is_leap(zeit[JAHR]))
			{
			if (zeit[TAG] <= 29) continue;
			}
		else
			{
			if (zeit[TAG] <= 28) continue;
			}
		}

	else											// die brigen Monate
		{
		if (zeit[TAG] <= pgm_read_byte(&days_of_month[zeit[MONAT]])) continue;
		}

	zeit[TAG] = 1;								// der letzte Tag des Vormonats war erreicht
	zeit[MONAT]++;
	if (zeit[MONAT] > 12)
		{
		zeit[MONAT] = 1;
		zeit[JAHR]++;
		}
	} while(--x);								// den Rundgang fr jede Minute wiederholen
}

// -----------------------------------------------------------------------------
// kopiert ein Zeit-Array
void DCF_copy_time(uint8_t *target, uint8_t *source)
{
register uint8_t i = 6;							// Element mit Index 5 ... 0 kopieren
do	{
	i--;
	target[i] = source[i];
	} while(i);
}

// -----------------------------------------------------------------------------
// vergleicht die Bytes der Arrays time1 und time2 byteweise miteinander
// liefert: TRUE  wenn identisch,
//          False wenn ungleich
uint8_t DCF_compare_time(uint8_t *time1, uint8_t *time2)
{
uint8_t i = 6;										// Element mit Index 5 ... 0 vergleichen
uint8_t r = TRUE;
do {
	i--;												// ein einziger fehlerhafter Vergleich reicht aus
	if (! (time1[i] == time2[i]) ) r = FALSE;
	} while (i);
return(r);
}

// -----------------------------------------------------------------------------
// Wird sinnvollerweise jeweils zu einer vollen Stunde aufgerufen, da nur dann ein Wechsel
// zwischen Sommer- und Winterzeit stattfinden kann.
//
// Prfen anhand von Datum und Zeit der internen Uhr, ob zwischen Sommer- und Winterzeit umgeschaltet werden muss.
//
// Am letzten Sonntag im Mrz um 02.00h wird die Uhr auf 03.00h,
// am letzten Sonntag im Oktober um 03.00h wird die Uhr auf 02.00h gestellt.
// Ausgehend von der aktuellen Sommerzeit wird geprft,
// wenn Sommerzeit = 0, ob Monat  = 3, Wochentag = 7 und Datum zwischen 25 und 31 und Stunde = 2
// wenn Sommerzeit = 1, ob Monat = 10, Wochentag = 7 und Datum zwischen 25 und 31 und Stunde = 3
// zur Kontrolle/Besttigung knnte noch das Ankndigungsbit A1 (Bit 16) herangezogen werden ...
// kritisch ist Umschaltung im Oktober, da hier der Zeitraum von 0.200h bis 03.00h zweimal nacheinander
// durchlaufen wird -> hier die aktuelle Einstellung der Sommerzeit mageblich:
// wenn Sommerzeit, dann Umschaltung, wenn (schon) Winterzeit, dann keine Umschaltung !
void switch_summertime(uint8_t *zeit)
{
if (zeit[TAG] > 24)							// Gewechselt wird zwischen dem 25. und dem 31. (Mrz bzw. Oktober)
	{
	if ((zeit[DOW] & 0x40) )				// nur an einem Sonntag (dann ist Bit.6 gesetzt)
		{
		if (zeit[MONAT] == 3)				// im Mrz
			{
			if (zeit[STUNDE] == 2)			// um 2:00h die Uhr auf 3:00h vorstellen
				{
				zeit[STUNDE] = 3;				// Zeit vorstellen auf 3;00h
				set_summertime(zeit, TRUE);// auf Sommerzeit schalten
				}
			}
													// im Oktober die Uhr zurckstellen (sofern noch Sommerzeit !)
													// 03.00h wird zweimal durchlaufen,
													// beim 2. Mal ist aber bereits auf Winterzeit umgeschaltet,
													// daher wird die Uhr nicht noch einmal zurckgestellt
		else if ((zeit[MONAT] == 10) && (is_summertime(zeit)))
			{
			if (zeit[STUNDE] == 3)
				{
				zeit[STUNDE] = 2;				// Zeit zurckstellen auf 2:00h
				set_summertime(zeit, FALSE);// und auf Winterzeit stellen
				}
			}

		}
	}
}

// ----------------------------------------------------------------------------
// Prft, ob das DCF-Signal Sommerzeit signalisiert
// Rckgabe: 1 = Sommerzeit
//           0 = Keine Sommerzeit
// es wird nicht DCF_zbits ausgewertet, sondern das Bit.7 in time[DOW] !
uint8_t is_summertime(uint8_t *zeit)
{
return(zeit[DOW] & 0x80); 					// Bit.7 in DOW signalisiert die Sommerzeit
}

// -----------------------------------------------------------------------------
// setzt/lscht das Sommerzeitflag in time[DOW].7
void set_summertime(uint8_t *zeit, uint8_t i)
{
if (i) zeit[DOW] |= 0x80; else zeit[DOW] &= ~(0x80);
}

// -----------------------------------------------------------------------------
// ein Schaltjahr liegt vor, wenn die Jahreszahl durch 4, 100 (oder 400) ohne Rest teilbar
// es werden nur die beiden Flle mod 4/ mod 100 untersucht

uint8_t is_leap(uint8_t j)
{													// Es liegt ein Schaltjahr vor,
if (!(j & 0x03)) return(TRUE);			// wenn mod 4 keinen Rest ergibt

while (j > 100) j -= 100;					// diese berprfung ist hier wohl eher berflssig ..
if (!(j)) return(TRUE);						// wenn mod 100 keinen Rest ergibt

return(FALSE);
}

// ----------------------------------------------------------------------------
// Vergleicht mit den letzten beiden Minuten
uint8_t DCF_time_valid(void)
{
return( DCF_compare_time(&time[BK2][0], &time[BK1][0]) && DCF_compare_time(&time[BK1][0], &time[DCF][0]) );
}


// ----------------------------------------------------------------------------
// legt eine Kopie der letzten beiden Minuten ab und incrementiert die Zeit so,
// dass ein direkter Abgleich mit der neuen Minute mglich ist
// Zuerst Vergleichen, dann Sichern
void DCF_backup(void)
{
DCF_copy_time(&time[BK2][0], &time[BK1][0]);
DCF_copy_time(&time[BK1][0], &time[DCF][0]);

DCF_incr_time(&time[BK2][0],1);
DCF_incr_time(&time[BK1][0],1);
}

// ----------------------------------------------------------------------------
// resetet alle Bits
// und (das ist das wichtigste !) setzt den Index auf 0 zurck
void DCF_reset_bits(void)
{
register uint8_t i = 8;		// Byte 7 .. 0 kopieren und dann lschen

do	{
	dcf_bits[--i] = 0;		// Inhalt lschen
	} while (i);
bit_idx = 0;					// den Index fr die nchste Minute zurcksetzen
}

// ----------------------------------------------------------------------------
// setzt den Bit-Wert von DCF_in in die nchste frei Bitposition
// incrementiert den Index bit_idx, solange dieser < 63
// setzt das Flag DCF_RX_58, wenn genau 58 Bit empfangen worden sind.
// werden mehr als 58 Bit empfangen, wird das Fehlerflag ERR_BIT_CNT1 gesetzt

void DCF_save_bit(uint8_t v)
{
uint8_t i,j;

j = bit_idx / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit_idx % 8;			// Bitposition
								// Index auf die folgende Position setzen
								// berlauf ber Arraygrenzen verhindern
if (bit_idx < 59) bit_idx++; else error |= (1<<ERR_BIT_CNT2);

if (v) dcf_bits[j] |= (1<<i); else dcf_bits[j] &= ~(1<<i);

if (bit_idx == 59) flag |= (1<<DCF_RX_58);

}

// ----------------------------------------------------------------------------
// holt in einem Array von 8 Byte das Bit x (0..63)
uint8_t DCF_get_bit(uint8_t bit)
{
uint8_t i, j, r;

j = bit / 8;			// das zustndige Byte aus dem Index ermitteln
i = bit % 8;			// Bitposition

r = (dcf_bits[j] & (1<<i));
if (r) r = 1;
return(r);
}

// ----------------------------------------------------------------------------
// liefert die Anzahl der empfangenen DCF-Bits zurck
uint8_t DCF_bits_received(void)
{
return(bit_idx - 1);
}

#ifdef SoD
// ----------------------------------------------------------------------------
// liefert die UhrZeit als Anzahl der abgelaufenen Sekunden des Tages
// Da 24 * 60 * 60 = 86.400 > 65536, wird nur jede 2. Sekunde gezhlt ! 
uint16_t DCF_Get_SoD(void)
{
uint16_t n;
n  = time[CLK][STUNDE] *  60;					// Stunden in Minuten umrechnen
n += time[CLK][MINUTE];							// Minuten addieren
n *= 30;												// alles  mit 30 Sek/Min multiplizieren
n += time[CLK][SEKUNDE] / 2;					// Halbe Anzahl der Sekunden addieren
return(n);
}
#endif

#ifdef DoY
// -----------------------------------------------------------------------------
// Ermittelt die lfd. Nummer des aktuellen Tages im Jahr unter Bercksichtigung von Schaltjahren
uint16_t DCF_Get_DoY(void)
{
uint16_t z = 0;
uint8_t i = 1;

while (i < time[CLK][MONAT])					// die vollen Tage der Monate addieren
	{
	z+= pgm_read_byte(&days_of_month[i++]);
	if (i == 2)										// im Februar das Schaltjahr bercksichtigen
		{
		if (is_leap(time[CLK][JAHR])) z++;	// und einen Tag addieren fr den 29. Februar
		}
	}
z += time[CLK][TAG];								// nun noch die Tage des laufenden Monats addieren
return(z);
}

#endif

// --- [ eof ] ----------------------------------------------------------------
